/*
 * fpga.c
 *
 * Copyright (C) 2006 Luotao Fu <lfu@pengutronix.de>, Pengutronix
 * Copyright (C) 2008 Juergen Beisert <kernel@pengutronix.de>, Pengutronix
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the version 2 of the GNU General Public License
 * as published by the Free Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * Note:
 * The external FPGA device is an Altera Cyclone II, type EP2C8F256C8N
 *
 * It can be feed by an external serial flash device or by some GPIOs from
 * CPU side.
 *
 * Loading from external serial flash at reset time:
 * - in this case PSC6 can be used as an additional UART.
 * - JP11 must be open and R31 assembled
 * - JP8 must be closed at 1-2 to select "Active Serial Mode"
 * - U18 (EPCS4N/25P20) must be assembled
 *
 * Loading from the CPU at runtime:
 * - in this case PSC6 cannot be used as a UART, or if possible, not routed
 *   to the external pins while loading the FPGA
 * - in the case J11 is closed at 1-2
 *    - GPIO6 from CPU must be high
 * - in the case J11 is closed at 3-2
 *    - GPIO6 from CPU is free for alternative usage
 * - J8 must be closed at 2-3 to select "Passive Serial Mode"
 *
 * Now its possible to control:
 *
 * FPGA's             CPUs
 *  pin     through    pin
 * ---------------------------
 *  DCLK (in)        PSC6_3 (out)
 *  DATA0 (in)       PSC6_2 (out)
 *  CONFIG# (in)     PSC6_1 (out)
 *  DONE (out)       PSC6_0 (in)
 *  STATUS (out)     GPIO7 (in)
 */

#define DEBUG

#include <linux/init.h>
#include <linux/delay.h>

#include <linux/firmware.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/fs.h>
#include <linux/of.h>
#include <linux/of_platform.h>

#include <asm/io.h>
#include <asm/mpc52xx.h>
#include <asm/machdep.h>

#define TIMEOUT 50000

static char *firmware = NULL;
module_param(firmware, charp, 0400);
MODULE_PARM_DESC(firmware, "firmware filename");

static struct mpc52xx_gpio __iomem *gpio;
static struct mpc52xx_gpio_wkup __iomem *wkup;

/* this function is gone in the generic mpc5200 support */
static void __iomem *mpc52xx_find_and_map(const char *compatible)
{
	struct device_node *ofn;
	const u32 *regaddr_p;
	u64 regaddr64, size64;

	ofn = of_find_compatible_node(NULL, NULL, compatible);
	if (!ofn)
		return NULL;

	regaddr_p = of_get_address(ofn, 0, &size64, NULL);
	if (!regaddr_p) {
		of_node_put(ofn);
		return NULL;
	}

	regaddr64 = of_translate_address(ofn, regaddr_p);

	of_node_put(ofn);

	return ioremap((u32)regaddr64, (u32)size64);
}

static int map_resources(void)
{
	gpio = mpc52xx_find_and_map("mpc5200-gpio");
	wkup = mpc52xx_find_and_map("mpc5200-gpio-wkup");

	if (!gpio || !wkup) {
		pr_err("%s: error while mapping GPIO/GPIO_WKUP\n", __FUNCTION__);
		return -1;
	}

	return 0;
}

static int unmap_resources(void)
{
	if (gpio) {
		iounmap(gpio);
		gpio = NULL;
	}
	if (wkup) {
		iounmap(wkup);
		wkup = NULL;
	}

	return 0;
}

static void set_dclk(void)
{
	u32 data;

	data = in_be32(&gpio->simple_dvo);
	data |= 0x20000000;
	out_be32(&gpio->simple_dvo, data);
}

static void clear_dclk(void)
{
	u32 data;

	data = in_be32(&gpio->simple_dvo);
	data &= ~0x20000000;
	out_be32(&gpio->simple_dvo, data);
}

static void set_data(void)
{
	u32 data;

	data = in_be32(&gpio->simple_dvo);
	data |= 0x10000000;
	out_be32(&gpio->simple_dvo, data);
}

static void clear_data(void)
{
	u32 data;

	data = in_be32(&gpio->simple_dvo);
	data &= ~0x10000000;
	out_be32(&gpio->simple_dvo, data);
}

static void set_config(void)
{
	u8 config;

	config = in_8(&wkup->wkup_dvo);
	config |= 0x20;
	out_8(&wkup->wkup_dvo, config);
}

static void clear_config(void)
{
	u8 config;

	config = in_8(&wkup->wkup_dvo);
	config &= ~0x20;
	out_8(&wkup->wkup_dvo, config);
}

static int read_done(void)
{
	return (!!(in_8(&wkup->wkup_ival) & 0x10));
}

static int read_status(void)
{
	return (!!(in_8(&wkup->wkup_ival) & 0x80));
}

static u32 port_config_backup;
static u32 simple_gpioe_backup;
static u32 simple_ddr_backup;
static u8 wkup_gpioe_backup;
static u8 wkup_ddr_backup;

static void __devinit init_gpio_fpga(void)
{
	u32 port_config, simple_gpioe, simple_ddr;
	u8 wkup_gpioe, wkup_ddr, wkup_dvo;

	/* set Port to configure IRDA as GPIO */
	port_config_backup = port_config = in_be32(&gpio->port_config);
	port_config &= ~0x00700000;
	out_be32(&gpio->port_config, port_config);

	/* enabling GPIO_IRDA_0, GPIO_IRDA_1 */
	simple_gpioe = simple_gpioe_backup = in_be32(&gpio->simple_gpioe);
	simple_gpioe |= 0x30000000;
	out_be32(&gpio->simple_gpioe, simple_gpioe);

	/* set GPIO_IRDA_0 as out(Bit 2 low), GPIO_IRDA_1 as out(Bit 3 high)*/
	simple_ddr = simple_ddr_backup = in_be32(&gpio->simple_ddr);
	simple_ddr |= 0x30000000;
	out_be32(&gpio->simple_ddr, simple_ddr);

	/* enabling GPIO_WKUP_4, GPIO_WKUP_5, GPIO_WKUP_6, GPIO_WKUP_7 */
	wkup_gpioe = wkup_gpioe_backup = in_8(&wkup->wkup_gpioe);
	wkup_gpioe |= 0xF0;
	out_8(&wkup->wkup_gpioe, wkup_gpioe);

	/* set GPIO_WKUP_7, GPIO_4 as in (Bit 0 and Bit 3 low) and GPIO_WKUP_5,
	 * GPIO_WKUP_6 as out (Bit 1 and 2 high) */
	wkup_ddr = wkup_ddr_backup = in_8(&wkup->wkup_ddr);
	wkup_ddr &= ~0x90;
	wkup_ddr |= 0x60;
	out_8(&wkup->wkup_ddr, wkup_ddr);

	/* switch the multiplexer to drive the FPGA lines instead the UART ones */
	wkup_dvo = in_8(&wkup->wkup_dvo);
	wkup_dvo |= 0x40;	/* GPIO6 aka GPIO_WKUP_6 */
	out_8(&wkup->wkup_dvo, wkup_dvo);

	/* set config(GPIO_WKUP_5) to low and wait 50 us till the  */
	/* initialization of the datatransfer */
	clear_config();
	udelay(50);
	/*set config to high and wait till device ready (high on status)*/
	set_config();
}

static void __devinit exit_gpio_fpga(void)
{
	u8 wkup_dvo;

	/* switch the multiplexer to drive the UART lines instead the FPGA ones */
	wkup_dvo = in_8(&wkup->wkup_dvo);
	wkup_dvo &= ~0x40;	/* GPIO6 aka GPIO_WKUP_6 */
	out_8(&wkup->wkup_dvo, wkup_dvo);

	out_8(&wkup->wkup_ddr, wkup_ddr_backup);
	out_8(&wkup->wkup_gpioe, wkup_gpioe_backup);
	out_be32(&gpio->simple_ddr, simple_ddr_backup);
	out_be32(&gpio->simple_gpioe, simple_gpioe_backup);
	out_be32(&gpio->port_config, port_config_backup);

}

static int __devinit fpga_loader_probe(struct of_device *ofdev,
			const struct of_device_id *match)
{
	int bits_transferred = 0, size, timeout = TIMEOUT, err;
	const struct firmware *fw;
	u8 conf_done, status;
	u8 *buf;

	/* there must be a file name given! */
	if (!firmware) {
		dev_err(&ofdev->dev, "No firmware file specified!\n");
		return -EINVAL;
	}

	err = map_resources();
	if (err != 0) {
		dev_err(&ofdev->dev, "Cannot map GPIO resources\n");
		err = -ENODEV;
		goto on_error;
	}

	init_gpio_fpga();

	err = request_firmware(&fw, firmware, &ofdev->dev);
	if (err){
		dev_err(&ofdev->dev, "Cannot not load firmware!\n");
		goto on_error_1;
	}

	buf = fw->data;
	size = fw->size;

	while (1) {
		status = read_status();
		if (status)
			break;
		if (timeout <= 0) {
			err = -ENODEV;
			goto on_error_2;
		}
		timeout--;
		udelay(1);
	}

	dev_info(&ofdev->dev, "flashing firmware, %d bytes to go.\n", size);

	while (1) {
		conf_done = read_done();
		if (conf_done)
			break;

		if (*buf & 0x1)
			set_data();
		else
			clear_data();

		set_dclk();
		clear_dclk();

		*buf >>= 1;

		if (!(++bits_transferred % 8))
			buf++;

#ifdef DEBUG
		if(!(bits_transferred % 8192))
			printk(".");
#endif
		if (bits_transferred > size * 8){
			dev_warn(&ofdev->dev, "\nwarning: bit overrun\n");
			break;
		}
	}
	dev_dbg(&ofdev->dev, "\ndone\n");
	dev_dbg(&ofdev->dev, "config done status: %i\n", conf_done ? 1:0);
	dev_dbg(&ofdev->dev, "transferred bits: %i\n", bits_transferred);

on_error_2:
	release_firmware(fw);
on_error_1:
	exit_gpio_fpga();
on_error:
	unmap_resources();

	return err;
}

static int __devexit fpga_loader_remove(struct of_device *ofdev)
{
	return 0;
}

/* Device tree match table for this device */
static struct of_device_id fpga_ids[] = {
	{ .compatible = "phytec,fpga" },
	{}
};

static struct of_platform_driver fpga_driver = {
	.name = "fpga",
	.match_table = fpga_ids,
	.probe = fpga_loader_probe,
	.remove = __devexit_p(fpga_loader_remove),
};

static int __init fpga_init(void)
{
	int err;

	pr_debug("Registering FPGA firmware refresher\n");

	err = of_register_platform_driver(&fpga_driver);
	if (err != 0)
		pr_err("Unable to register fpga firmware refresher\n");

	return err;
}

static void __exit fpga_exit(void)
{
	of_unregister_platform_driver(&fpga_driver);
}

module_init(fpga_init);
module_exit(fpga_exit);
MODULE_AUTHOR("Luotao Fu, Juergen Beisert");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("FPGA Loader on pcm032 Board");
